/**
 * Created with IntelliJ IDEA.
 * User: aaronlin
 * Date: 2018/6/15
 * Time: 18:00
 */
import React, { Component } from "react"
import {
    StyleSheet,
    Text,
    View,
    ViewPropTypes,
} from "react-native";

import PropTypes from "prop-types";

import {
    EventEmitter,
} from '../RNExt';

/**
 *
 * @type {EventEmitter}
 */
const badgeEventEmitter = new EventEmitter();

/**
 *
 * @type {string}
 */
const MESSAGE_COUNT_EVENT = 'message';

export default class Badge extends Component {

    static propTypes = {
        ...ViewPropTypes,
        eventName: PropTypes.string,    //消息个数事件名称
        messageCount: PropTypes.number, //消息个数
        showTxt: PropTypes.bool //是否显示消息个数
    }

    static defaultProps = {
        messageCount:0,
        showTxt: false,
    }

    /**
     * 发送消息个数事件
     * @param pullRefreshEmitter
     */
    static sendMessageEvent(messageCount : number) : void{
        badgeEventEmitter.emit(MESSAGE_COUNT_EVENT, messageCount);
    }

    /**
     * 发送消息个数事件
     * @param pullRefreshEmitter
     */
    static sendCustomMessageEvent(eventName : string, messageCount : number) : void{
        badgeEventEmitter.emit(eventName, messageCount);
    }

    constructor(props) {
        super(props);

        const { eventName = MESSAGE_COUNT_EVENT, messageCount, showTxt } = this.props;
        badgeEventEmitter.addListener(eventName, this._messageCountListener);

        this.state = {
            messageCount: messageCount,
            showTxt: showTxt,
        };
    }

    /**
     * 消息个数
     * @param messageCount
     * @private
     */
    _messageCountListener = (messageCount : number = 0)=>{
        this.setState({ messageCount });
    }

    componentWillUnmount() {
        const { eventName = MESSAGE_COUNT_EVENT } = this.props;
        badgeEventEmitter.removeListener(eventName, this._messageCountListener);
    }

    render() {

        const { style, txtStyle, ...otherProps } = this.props;
        const { messageCount, showTxt } = this.state;

        if(messageCount <= 0){
            return null;
        }

        return (
            <View
                style={[styles.container, style]}
                pointerEvents={'none'}
                {...otherProps}
            >
                {
                    showTxt ?
                        <Text style={[styles.txt, txtStyle]}>{messageCount}</Text>
                        :
                        null
                }

            </View>
        );
    }

}

const styles = StyleSheet.create({
    container: {
        position:'absolute',
        top:0,
        right:0,
        width:8,
        height:8,
        backgroundColor:'red',
        borderRadius:4,
        justifyContent:'center',
        alignItems:'center',
        overflow:'hidden',
    },
    txt:{
        color:'black',
        fontSize:12,
    },
});