/**
 * Created with IntelliJ IDEA.
 * User: aaronlin
 * Date: 2018/5/7
 * Time: 13:26
 */

import React, { Component, PureComponent } from 'react';
import {
    View,
    Text,
    Dimensions,
    PanResponder,
    Animated,
} from 'react-native';

import PropTypes from 'prop-types';

import BaseAdapterComponent from '../base/BaseAdapterComponent';

/**
 *
 */
const { width:SCREEN_WIDTH, height:SCREEN_HEIGHT } = Dimensions.get('window');

/**
 * 最小 X 的移动距离
 * @type {number}
 */
const MIN_MOVE_X_DISTANCE = 0.66;
/**
 * 最小 X 的移动距离
 * @type {number}
 */
const MAX_MOVE_Y_DISTANCE = 0.66;
/**
 * 最小触发动画的距离
 * @type {number}
 */
const MIN_DISTANCE = SCREEN_WIDTH/10;
/**
 *
 * @type {number}
 */
const ANIM_DURATION = 150;
/**
 *
 * @type {number}
 */
const AUTO_PLAY_DURATION = 3000;

/**
 *
 * @type {{item0: number, item1: Object.width, item2: number}}
 */
const ITEM_TRANSLATEX_INIT = {
    translateX0:0,
    translateX1:SCREEN_WIDTH,
    translateX2:SCREEN_WIDTH*2,
}

/**
 * 复用的 ViewPager
 */
export default class ViewPager extends PureComponent {

    /**
     *
     * @type {{dataList: *, component: *, autoPlay: React$PropType$Primitive.<boolean>, dotVisible: React$PropType$Primitive.<boolean>}}
     */
    static propTypes = {
        ...View.propTypes,
        /**
         * 数据数组
         */
        dataList:PropTypes.array.isRequired,
        /**
         * 要渲染的 BaseAdapterComponent 组件
         */
        component:PropTypes.func.isRequired,
        /**
         * 自动播放
         */
        autoPlay:PropTypes.bool,
        /**
         * 底部的 dot 指示器 是否显示
         */
        dotVisible:PropTypes.bool,
        /**
         *
         */
        itemRight:PropTypes.number,
        /**
         *
         */
        itemLeft:PropTypes.number,
    }

    /**
     *
     * @type {{dataList: Array}}
     */
    static defaultProps = {
        dataList:[],
        autoPlay:true,
        dotVisible:true,
        itemLeft:0,
        itemRight:0,
    }

    /**
     * itemView 宽度
     * @type {Object.width}
     * @private
     */
    _itemWidth = SCREEN_WIDTH;

    /**
     *
     * @type {null}
     * @private
     */
    _panResponder = null;

    /**
     *
     * @type {Array}
     * @private
     */
    _initDataArr = [];

    /**
     *
     * @type {{}}
     * @private
     */
    _itemTranslateX = ITEM_TRANSLATEX_INIT;

    /**
     * 是否正在动画
     * @type {boolean}
     * @private
     */
    _isAniming = false;

    /**
     *
     * @type {null}
     * @private
     */
    _animTimeoutId = null;

    /**
     *
     * @type {number}
     * @private
     */
    _currPosition = 0;

    constructor(props) {
        super(props);

        const { component } = this.props;

        if(!component.__proto__ || component.__proto__ !== BaseAdapterComponent){
            throw new Error('ViewPager.component must be extends BaseAdapterComponent');
        }

        this._init();

        this.state = {
            dataList: this.props.dataList,
            itemAnim:{
                translateX0: new Animated.Value(this._itemTranslateX.translateX0),
                translateX1: new Animated.Value(this._itemTranslateX.translateX1),
                translateX2: new Animated.Value(this._itemTranslateX.translateX2),
            }
        };

        this._moveItemTranslate();

        this._initComponentData();

        this._initPanResponder();
    }

    componentWillReceiveProps(nextProps){
        if(nextProps){
            let { dataList } = nextProps;
            this._endAllAnim();

            this.setState({ dataList }, this._startAutoAnim);
        }
    }

    /**
     *
     */
    componentDidMount() {
        this._startAutoAnim();
    }

    /**
     *
     */
    componentWillUnmount() {
        this._endAllAnim();
    }

    /**
     *
     * @private
     */
    _init = ()=>{
        const { itemLeft, itemRight } = this.props;
        this._itemWidth = SCREEN_WIDTH - itemLeft - itemRight;
        this._itemTranslateX.translateX1 = this._itemWidth;
        this._itemTranslateX.translateX2 = this._itemWidth * 2;
    }

    /**
     *
     * @private
     */
    _initPanResponder = ()=>{
        this._panResponder = PanResponder.create({
            // 要求成为响应者：
            onStartShouldSetPanResponder: (evt, gestureState) => false,
            onStartShouldSetPanResponderCapture: (evt, gestureState) => false,
            onMoveShouldSetPanResponder: (evt, gestureState) => {
                // console.log('onMoveShouldSetPanResponder===========> ', gestureState.dy);
                const { dx, dy } = gestureState;
                return !this._isAniming && Math.abs(dx)>=MIN_MOVE_X_DISTANCE;
            },
            onMoveShouldSetPanResponderCapture: (evt, gestureState) => {
                // console.log('onMoveShouldSetPanResponderCapture===========> ', gestureState.dy);
                const { dx, dy } = gestureState;
                return !this._isAniming && Math.abs(dx)>=MIN_MOVE_X_DISTANCE;
            },
            // 开始手势操作。给用户一些视觉反馈，让他们知道发生了什么事情！
            // gestureState.{x,y} 现在会被设置为0
            onPanResponderGrant: this._onPanResponderGrant,
            // 最近一次的移动距离为gestureState.move{X,Y}
            // 从成为响应者开始时的累计手势移动距离为gestureState.d{x,y}
            onPanResponderMove: this._onPanResponderMove,
            //监视器被要求终止，返回false表示不同意终止，或者不处理
            onPanResponderTerminationRequest: this._onPanResponderTerminationRequest,
            // 用户放开了所有的触摸点，且此时视图已经成为了响应者。
            // 一般来说这意味着一个手势操作已经成功完成。
            onPanResponderRelease: this._onPanResponderRelease,
            // 另一个组件已经成为了新的响应者，所以当前手势将被取消。
            onPanResponderTerminate: this._onPanResponderRelease,
            // 返回一个布尔值，决定当前组件是否应该阻止原生组件成为JS响应者
            // 默认返回true。目前暂时只支持android。
            onShouldBlockNativeResponder: (evt, gestureState) => {
                return true;
            },
        });
    }

    /**
     * 初始化组件数据
     * @private
     */
    _initComponentData = ()=>{
        this._initDataArr = [];
        const { dataList } = this.state;

        let countTmp = dataList.length;
        if(countTmp === 2){
            countTmp = 3;
        }

        for(let index = 0; index<countTmp; index++){
            if(index >=3){
                break;
            }

            let position = index;
            if(position > dataList.length - 1){
                position = dataList.length - 1;
            }

            let initData = {
                ref: 'component-translateX'+index,
                param: {
                    position:position,
                    data:dataList[position]
                }
            }

            this._initDataArr.add(initData);
        }
    }

    /**
     *
     * @private
     */
    _initItemTranslateX = ()=>{
        const { itemAnim } = this.state;

        this._itemTranslateX = {
            translateX0:itemAnim.translateX0._value,
            translateX1:itemAnim.translateX1._value,
            translateX2:itemAnim.translateX2._value,
        }
    }

    /**
     *
     * @param evt
     * @param gestureState
     * @private
     */
    _onPanResponderGrant = (evt, gestureState) => {
        if(this._isAniming){
            return ;
        }
        // console.log('onPanResponderGrant===========> ');

        this._initItemTranslateX();
    }

    /**
     *
     * @param evt
     * @param gestureState
     * @private
     */
    _onPanResponderMove = (evt, gestureState) => {
        if(this._isAniming){
            return ;
        }
        this._endAutoAnim();

        // console.log('onPanResponderMove=======> ', gestureState);
        let { dx } = gestureState;

        const { itemAnim } = this.state;

        for(let key in itemAnim){
            itemAnim[key].setValue(this._itemTranslateX[key] + dx);
        }
    }

    /**
     *
     * @param evt
     * @param gestureState
     * @returns {boolean}
     * @private
     */
    _onPanResponderTerminationRequest = (evt, gestureState) => {
        this._startAnim();
        return true;
    }

    /**
     *
     * @param evt
     * @param gestureState
     * @private
     */
    _onPanResponderRelease = (evt, gestureState) => {
        // console.log('onPanResponderRelease================>');

        this._startAnim();
    }

    /**
     * 移动 item Translate
     * @private
     */
    _moveItemTranslate = ()=>{
        let currKey = this._getCurrItemKey();
        if(currKey){
            const { itemAnim } = this.state;

            if(currKey === 'translateX0'){
                itemAnim.translateX2.setValue(-this._itemWidth);
                itemAnim.translateX1.setValue(this._itemWidth);
            }else if(currKey === 'translateX1'){
                itemAnim.translateX0.setValue(-this._itemWidth);
                itemAnim.translateX2.setValue(this._itemWidth);
            }else if(currKey === 'translateX2'){
                itemAnim.translateX1.setValue(-this._itemWidth);
                itemAnim.translateX0.setValue(this._itemWidth);
            }

            this._initItemTranslateX();
        }
    }

    /**
     *
     * @private
     */
    _getCurrItemKey = ()=> {
        const { itemAnim } = this.state;
        for (let key in itemAnim) {
            let value = itemAnim[key]._value;
            if (value === 0) {
                return key;
            }
        }
    }

    /**
     * 通知更新组件
     * @private
     */
    _notifyUpdateComponent = ()=>{
        let currKey = this._getCurrItemKey();
        if(currKey){
            const { dataList } = this.state;

            let preRef, nextRef, prePosition, nextPosition;
            if(currKey === 'translateX0'){
                preRef = 'component-translateX2';
                nextRef = 'component-translateX1';
            }else if(currKey === 'translateX1'){
                preRef = 'component-translateX0';
                nextRef = 'component-translateX2';
            }else if(currKey === 'translateX2'){
                preRef = 'component-translateX1';
                nextRef = 'component-translateX0';
            }

            prePosition = this._currPosition - 1;
            nextPosition = this._currPosition + 1;
            if(prePosition < 0){
                prePosition = dataList.length - 1;
            }
            if(nextPosition >= dataList.length){
                nextPosition = 0;
            }

            this.refs[preRef] && this.refs[preRef].onUpdateComponent(prePosition, dataList[prePosition]);
            this.refs[nextRef] && this.refs[nextRef].onUpdateComponent(nextPosition, dataList[nextPosition]);
        }
    }

    /**
     * 动画之后改变 position
     * @param currValue
     * @private
     */
    _changePositionAfterAnim = (currValue : number)=>{
        const { dataList } = this.state;

        if(currValue < 0 && Math.abs(currValue) >= MIN_DISTANCE) {  //往左滑动
            if(this._currPosition >= dataList.length-1){
                this._currPosition = 0;
            }else{
                ++this._currPosition;
            }
            this._notifyUpdateComponent();
        }else if(currValue > 0 && currValue >= MIN_DISTANCE) {    //往右滑动
            if(this._currPosition <= 0){
                this._currPosition = dataList.length-1;
            }else{
                --this._currPosition;
            }
            this._notifyUpdateComponent();
        }

        this.refs.dotComp && this.refs.dotComp.setCountAndPosition(dataList.length, this._currPosition);
    }

    /**
     * 开始自动播放动画
     * @private
     */
    _startAutoAnim = ()=>{
        this._endAutoAnim();

        const { autoPlay } = this.props;
        if(autoPlay){
            this._animTimeoutId = setTimeout(this._startAnim, AUTO_PLAY_DURATION, true);
        }
    }

    /**
     * 结束自动播放的动画
     * @private
     */
    _endAutoAnim = ()=>{
        this._animTimeoutId && clearTimeout(this._animTimeoutId);
        this._animTimeoutId = null;
    }

    /**
     * 结束所有动画
     * @private
     */
    _endAllAnim = ()=>{
        this._endAutoAnim();

        const { itemAnim } = this.state;
        const anims = Object.values(itemAnim);
        for(let anim of anims){
            anim.stopAnimation();
        }
    }

    /**
     * 开始动画
     * @private
     */
    _startAnim = (isAutoAnim : boolean = false)=>{
        const { dataList } = this.state;
        if(this._isAniming || dataList.length<=1){
            return ;
        }
        const { itemAnim } = this.state;

        for(let key in this._itemTranslateX){
            let value = this._itemTranslateX[key];
            if(value === 0){
                let currValue = itemAnim[key]._value;

                if(currValue != 0 || isAutoAnim){
                    const animArr = [];

                    for(let itemAnimKey in itemAnim){
                        let itemAnimObj = itemAnim[itemAnimKey];
                        let fromValue = itemAnimObj._value;
                        let toValue;
                        if(currValue < 0 || isAutoAnim) {  //往左滑动
                            if(Math.abs(currValue) >= MIN_DISTANCE || isAutoAnim){
                                toValue = fromValue - (this._itemWidth - Math.abs(currValue));
                            }else{
                                toValue = this._itemTranslateX[itemAnimKey];
                            }
                        }else if(currValue > 0){    //往右滑动
                            if(currValue >= MIN_DISTANCE){
                                toValue = fromValue + (this._itemWidth - currValue);
                            }else{
                                toValue = this._itemTranslateX[itemAnimKey];
                            }
                        }
                        let anim = Animated.timing(itemAnimObj, {
                            fromValue: fromValue,
                            toValue: toValue,
                            duration: ANIM_DURATION,
                            isInteraction: false,
                        });

                        animArr.add(anim);
                    }

                    if(animArr.length > 0){
                        this._isAniming = true;
                        Animated.parallel(animArr).start(()=>{
                            this._changePositionAfterAnim(isAutoAnim ? -MIN_DISTANCE : currValue);
                            this._moveItemTranslate();
                            this._isAniming = false;

                            this._startAutoAnim();
                        });
                    }
                }
                break;
            }
        }
    }

    render(){
        this._endAllAnim();
        this._initComponentData();

        const { style, component:ItemComponent, dotVisible, itemLeft, itemRight } = this.props;
        const { dataList } = this.state;
        if(dataList.length === 0){
            return null;
        }

        return (
            <View {...this._panResponder.panHandlers} style={[styles.container, style]} pointerEvents="box-none">
                {
                    this._initDataArr.map((item, index)=>{

                        let backgroundColor = null;
                        if(index === 0){
                            backgroundColor = 'blue';
                        }else if(index === 1){
                            backgroundColor = 'black';
                        }else{
                            backgroundColor = 'red';
                        }
                        let itemStyle = {
                            // backgroundColor: backgroundColor,
                            left:itemLeft,
                            right:itemRight,
                            transform:[{translateX:this.state.itemAnim['translateX'+index]}]
                        }

                        const itemView = (
                            <Animated.View key={item.ref} style={[styles.itemContainer, itemStyle]} pointerEvents="box-none">
                                <ItemComponent ref={item.ref} {...item.param} />
                            </Animated.View>
                        );
                        return itemView;
                    })
                }

                <DotComponent ref={'dotComp'} visible={dotVisible} count={dataList.length} currPosition={this._currPosition}/>
            </View>
        );
    }
}

/**
 *
 */
class DotComponent extends Component{

    constructor(props) {
        super(props);

        const { count, currPosition, visible } = this.props;

        this.state = {
            count,
            currPosition,
            visible,
        };
    }

    /**
     * 设置当前
     */
    setCountAndPosition = (count : number, currPosition : number)=>{
        this.setState({ count, currPosition });
    }

    render(){
        const { count, currPosition, visible } = this.state;
        if (!visible || count <= 1)
            return null;

        let dots = [];
        const ActiveDot = <View style={[{
            backgroundColor: '#ffffff',
            width: 8,
            height: 8,
            borderRadius: 4,
            marginLeft: 3,
            marginRight: 3,
            marginTop: 3,
            marginBottom: 3
        }]} />
        const Dot = <View style={[{
            backgroundColor: 'rgba(0,0,0,.2)',
            width: 8,
            height: 8,
            borderRadius: 4,
            marginLeft: 3,
            marginRight: 3,
            marginTop: 3,
            marginBottom: 3
        }]} />
        for (let i = 0; i < count; i++) {
            dots.push(i === currPosition
                ? React.cloneElement(ActiveDot, {key: i})
                : React.cloneElement(Dot, {key: i})
            )
        }
        return (
            <View style={[styles.dotBottomContainer]} pointerEvents="none">
                {dots}
            </View>
        );
    }

}

const styles = {
    container:{
        position:'relative',
        flexDirection:'row',
        justifyContent:'center',
        alignItems:'center',
        width:SCREEN_WIDTH,
        backgroundColor:'transparent',
        aspectRatio:1.8,
        overflow:'hidden'
    },
    itemContainer:{
        position:'absolute',
        top:0,
        bottom:0,
        left:0,
        right:0,
    },
    dotBottomContainer: {
        position: 'absolute',
        bottom: 5,
        left: 0,
        right: 0,
        flexDirection: 'row',
        flex: 1,
        justifyContent: 'center',
        alignItems: 'center',
        backgroundColor: 'transparent'
    },
};